import calendar
import ctypes
import json
import os
import signal
from datetime import datetime
from multiprocessing.sharedctypes import RawArray, RawValue

import numpy as np
import sys
import time
import traceback
import multiprocessing
from multiprocessing import Process, Queue, Value, Lock, Semaphore

import psutil
import torch
from modules import math_utils, frame_32_utils
from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils
from modules.logging_utils import LoggingUtils, log_process_id
from modules.profiler_utils import profile_func
from modules_core.prototypes_data import Config
import matplotlib.pyplot as plt

FREE = 0
NOT_PROCESSED = 1
DONE = 2

class WorkerDenoiserExtractor(Process):

    def __init__(self, args, cpu_affinity=None, cuda_device=0):
        self.cpu_affinity = cpu_affinity
        if sys.platform != 'darwin':
            multiprocessing.set_start_method('spawn', force=True)
        super().__init__(daemon=False)

        self.args = args
        self.cuda_device = cuda_device

        if not hasattr(self.args, 'leading_zero_grad_for_filtering'):
            self.args.leading_zero_grad_for_filtering = 64

        self.args.device = 'cuda'
        self.config = Config()

        self.is_running = RawValue(ctypes.c_bool, True)
        self.is_initialized = RawValue(ctypes.c_bool, False)

        self.max_data_input_frames = 64
        self.len_data_input_frame = self.args.input_timesteps_max * self.args.nb_melspec_bins
        self.max_data_input = self.max_data_input_frames * self.len_data_input_frame

        self.lock_data_slot_state = Lock()

        # busy slots (busy == true)
        self.array_data_slot_state = RawArray(ctypes.c_byte, self.max_data_input_frames)

        # input & output reuse
        self.array_data_frames = RawArray(ctypes.c_float, self.max_data_input)
        self.array_data_is_silent = RawArray(ctypes.c_bool, self.max_data_input_frames)

        self.noise_model_percent = RawValue(ctypes.c_float, 0.5) # min noise percentage to count part as noise
        self.noise_model_parts_min = RawValue(ctypes.c_ushort, 2) # min number of parts to count as noise whole frame

        self.is_started = RawValue(ctypes.c_bool, False)
        self.is_signal_to_log_status = RawValue(ctypes.c_bool, False)

    def start(self):
        if sys.platform != 'darwin':
            multiprocessing.set_start_method('spawn', force=True)
        super().start()
        while not self.is_started.value:
            time.sleep(0.01)


    # shared function
    def stop(self):
        self.is_running.value = False

    @staticmethod
    def send_sync(workers_denoiser_extractors, frame_spec):
        slot_id = None
        is_running = True
        worker = None
        while is_running:
            for worker in workers_denoiser_extractors:
                is_running = worker.is_running.value
                slot_id = worker.set_input_data(
                    frame_spec
                )
                if slot_id is not None:
                    break
            if slot_id is not None:
                break
            else:
                time.sleep(1e-3)

        output = None

        while is_running:
            is_running = worker.is_running.value
            output = worker.get_output_data(slot_id)
            if output is not None:
                frame_denoised, frame_data_is_noise = output
                break
            else:
                time.sleep(1e-3)
        return output

    # shared function
    # if no slots available then return None
    def set_input_data(self, frame_spec):
        assert (frame_spec.shape[1] == self.args.nb_sfft_bins), f'frame_mel should have been {self.args.nb_sfft_bins} instead of {frame_spec.shape[1]}'

        slot_id = None

        try:
            for idx_slot in range(self.max_data_input_frames):
                self.lock_data_slot_state.acquire()
                if self.array_data_slot_state[idx_slot] == FREE:
                    slot_id = idx_slot

                    idx_input_start = idx_slot * self.len_data_input_frame
                    idx_input_end = (idx_slot + 1) * self.len_data_input_frame

                    self.array_data_frames[idx_input_start:idx_input_end] = frame_spec.flatten()[:]

                    self.array_data_slot_state[idx_slot] = NOT_PROCESSED
                try:
                    self.lock_data_slot_state.release()
                except:
                    pass
                if slot_id is not None:
                    break
        except Exception as e:
            LoggingUtils.exception(e)

        return slot_id

    # shared function
    # return None if not ready
    def get_output_data(self, slot_id):
        output = None
        try:
            if self.array_data_slot_state[slot_id] == DONE:
                idx_input_start = slot_id * self.len_data_input_frame
                idx_input_end = (slot_id + 1) * self.len_data_input_frame

                frame_speech = np.array(self.array_data_frames[idx_input_start:idx_input_end]).reshape((self.args.input_timesteps_max, self.args.nb_melspec_bins))
                frame_data_is_noise = self.array_data_is_silent[slot_id]
                output = (frame_speech, frame_data_is_noise)

                # release slot
                self.array_data_slot_state[slot_id] = FREE # mark as free
        except Exception as e:
            LoggingUtils.exception(e)


        return output

    def load_model(self, model_name, path_model_args):
        LoggingUtils.info(f'loading: {model_name}')

        path_model = path_model_args.replace('.json', '.pt')

        with open(path_model_args, 'r') as fp:
            args = json.load(fp)
            args = DictToObj(**args)
        args.tf_path_data_args_json = self.args.path_data_args

        Model = getattr(__import__('modules_ext.' + args.tf_model, fromlist=[model_name]), 'Model')
        model = Model(args)

        LoggingUtils.debug(
            f'self.cuda_device: {self.cuda_device} device_count:{torch.cuda.device_count()}'
        )

        self.args.tf_cuda = True
        self.args.device = 'cuda'
        model = model.cuda(self.cuda_device)
        LoggingUtils.info(f'{model_name} cuda loaded')

        LoggingUtils.info(f'{model_name} params load')

        model = model.eval()

        if self.args.tf_cuda:
            model.load_state_dict(torch.load(path_model, map_location=lambda storage, loc: storage.cuda(self.cuda_device)))
        else:
            model.load_state_dict(torch.load(path_model, map_location='cpu'))

        LoggingUtils.info(f'{model_name} loaded')
        LoggingUtils.info(f'Denoiser model {path_model_args} loaded')

        model = model.eval()  # very important call to avoid dropouts

        return model

    def init(self):
        self.lock_data_slot_state.acquire()
        try:
            for idx_slot in range(self.max_data_input_frames):
                self.array_data_slot_state[idx_slot] = FREE
        except Exception as e:
            LoggingUtils.exception(e)
        self.lock_data_slot_state.release()

        LoggingUtils.info(f'path_model_denoise_speech_args: {self.args.path_model_denoise_speech_args}')
        self.model_denoiser = self.load_model('Model', self.args.path_model_denoise_speech_args)

        # set initialized only after model loaded
        self.lock_data_slot_state.acquire()
        self.is_initialized.value = True
        self.lock_data_slot_state.release()

    def scale_frames(self, frames):
        # normalize frames within batch
        for frame_idx in range(frames.shape[0]):
            frame = frames[frame_idx]
            min_S = np.min(frame)
            max_S = np.max(frame)
            if max_S - min_S != 0:
                frame -= min_S
                frame /= (max_S - min_S)  # scale 0.0 to 1.0
            frames[frame_idx] = frame
        return frames

    def call_model_frame(self, np_frames, model):

        t_frame = torch.FloatTensor(np_frames)
        if self.args.tf_cuda:
            t_frame = t_frame.to(f'{self.args.device}:{self.cuda_device}')
        else:
            t_frame = t_frame.to(self.args.device)

        t_frame = model.forward(t_frame)
        np_frame_out = self.scale_frames(t_frame.to('cpu').data.numpy())

        return np_frame_out

    def is_silent_frame(self, sub_frames):

        sub_frames_count = []
        is_silence = False

        idx_y_start = int(sub_frames.shape[1] * self.config.silence_detection_spec_threshold_min)
        idx_y_end = int(sub_frames.shape[1] * self.config.silence_detection_spec_threshold_max)
        idx_x_end = int(sub_frames.shape[0] * self.config.silence_detection_sub_frames_percent)

        # already sub_frames min .. max => 0.0 .. 1.0

        if self.config.is_debug_silence_frames_save_img:
            sub_frames_colored = np.expand_dims(sub_frames, 2)
            sub_frames_colored = np.tile(sub_frames_colored, (1, 1, 3))  # expand to RGB as grayscale
            sub_frames_colored = sub_frames_colored ** 0.2

        max_sub_frame_list = []
        for idx in range(sub_frames.shape[0]):

            avg_sub_frames = sub_frames[idx, idx_y_start:idx_y_end]
            max_sub_frame = np.max(avg_sub_frames)
            max_sub_frame_list.append(max_sub_frame)

            if max_sub_frame < self.config.silence_detection_amplitude_threshold:
                if self.config.is_debug_silence_frames_save_img:
                    sub_frames_colored[idx, idx_y_start:idx_y_end, 0] = 1.0 # red channel

                sub_frames_count.append(idx)
                if len(sub_frames_count) >= idx_x_end:
                    is_silence = True
                    break
            if is_silence:
                break

        if self.config.is_debug_silence_frames:
            timestamp_debug = time.time()
            LoggingUtils.debug(
                f'timestamp_debug: {timestamp_debug} '
                f'is_silence: {is_silence} '
                f'sub_frames_count: {len(sub_frames_count)} / {idx_x_end} '                
                f'median: {round(np.median(max_sub_frame_list), 2)} max: {round(np.max(max_sub_frame_list), 2)}')

            if self.config.is_debug_silence_frames_save_img:
                plt.imsave(
                    f'./tmp/{"silent" if is_silence else "voice"}_{timestamp_debug}.png',
                   np.swapaxes(np.clip(sub_frames_colored, 0.0, 1.0), axis1=0, axis2=1)
                )

        return is_silence

    def denoise_frame(
            self,
            np_frames,
            model):

        np_frames = np.array(np_frames)
        frames_result = np.array(np_frames)
        frames_is_noise = [False] * np_frames.shape[0]

        for idx_iter in range(self.config.silence_denoiser_iterations):
            if idx_iter == 0:
                np_frames = self.scale_frames(np_frames)
            if idx_iter > 0:
                np_frames *= self.config.silence_denoiser_amplitude
            np_frames = self.call_model_frame(np_frames, model)
            if idx_iter == 0: # take resut from first iteration for voiceID
                frames_result = np.array(np_frames)

        # for silence detection might have more iterations
        for idx_frame in range(np_frames.shape[0]):
            frames_is_noise[idx_frame] = self.is_silent_frame(np_frames[idx_frame])
            if frames_is_noise[idx_frame]:
                frames_result[idx_frame] = np_frames[idx_frame]

        return frames_result, frames_is_noise


    def set_signal_to_log(self):
        self.is_signal_to_log_status.value = True

    def run(self):
        profile_func(self.run_inner, self.args, profile_name='denoiser_extractor')

    def load_config_json(self): # server specific
        with open(self.args.path_config, 'r') as f:
            FileUtils.lock_file(f)
            config_tmp = json.load(f)
            for key in config_tmp.keys():
                if not hasattr(self.config, key) or getattr(self.config, key) != config_tmp[key]:
                    if self.args.is_debug_config_values:
                        LoggingUtils.warn(f'WorkerDenoiserExtractor load_config_json {key}: {config_tmp[key]}')
                    setattr(self.config, key, config_tmp[key]) # ref to object cannot change as used in managers
            FileUtils.unlock_file(f)

    def run_inner(self):

        try:
            LoggingUtils.remove_handlers()
            LoggingUtils.create()

            p = psutil.Process()
            all_cpus = np.arange(psutil.cpu_count(), dtype=np.int).tolist()
            p.cpu_affinity(all_cpus)
            torch.set_num_threads(psutil.cpu_count())
            LoggingUtils.warn(f'Denoiser: {p.pid} cpu_affinity: {p.cpu_affinity()}')

            self.init()

            original_sigint_handler = signal.getsignal(signal.SIGINT)
            original_sigterm_handler = signal.getsignal(signal.SIGTERM)

            def func_terminate(signum, frame):
                signal.signal(signal.SIGINT, original_sigint_handler)
                signal.signal(signal.SIGTERM, original_sigterm_handler)
                LoggingUtils.info('terminate signal denoiser')
                self.is_running.value = False
                try:
                    self.lock_data_slot_state.release()
                except:
                    pass

            signal.signal(signal.SIGINT, func_terminate)
            signal.signal(signal.SIGTERM, func_terminate)

            LoggingUtils.warn('Denoiser loop initialized')
            self.is_started.value = True
            time_reload_config = time.time()

            while self.is_running.value:

                if time.time() - time_reload_config > 10.0:
                    time_reload_config = time.time()
                    self.load_config_json()

                time.sleep(1e-3) # very important because of GIL
                if self.is_signal_to_log_status:
                    self.is_signal_to_log_status = False
                    log_process_id('denoiser_extractor', 'signal', self.args)

                idx_slots = []
                frames = []
                frames_is_noise = []

                for idx_slot in range(self.max_data_input_frames):
                    if self.array_data_slot_state[idx_slot] == NOT_PROCESSED:
                        idx_slots.append(idx_slot)

                        idx_input_start = idx_slot * self.len_data_input_frame
                        idx_input_end = (idx_slot + 1) * self.len_data_input_frame

                        frame_denoised = np.array(self.array_data_frames[idx_input_start:idx_input_end]).reshape((self.args.input_timesteps_max, self.args.nb_melspec_bins))
                        frames.append(frame_denoised)


                if len(idx_slots) > 0:
                    frames, frames_is_noise = self.denoise_frame(
                        frames,
                        self.model_denoiser,
                    )

                    for idx_queue, idx_slot in enumerate(idx_slots):

                        idx_input_start = idx_slot * self.len_data_input_frame
                        idx_input_end = (idx_slot + 1) * self.len_data_input_frame

                        self.array_data_frames[idx_input_start:idx_input_end] = frames[idx_queue].flatten()
                        self.array_data_is_silent[idx_slot] = frames_is_noise[idx_queue]

                        # mark finished
                        self.array_data_slot_state[idx_slot] = DONE

        except Exception as e:
            LoggingUtils.exception(e)
        log_process_id('denoiser_extractor', 'signal', self.args)
        LoggingUtils.warn('exit denoiser_extractor.run')
