import ctypes
import json
import os
import pickle
import signal
from collections import OrderedDict
from datetime import datetime
from multiprocessing.sharedctypes import RawArray, RawValue

import numpy as np
import sys
import time
import traceback
import multiprocessing
from multiprocessing import Process, Queue, Value, Lock, Semaphore
from typing import Tuple, List

import psutil
import torch
import torch.nn
from modules.dict_to_obj import DictToObj
from modules.logging_utils import LoggingUtils, log_process_id
from modules.profiler_utils import profile_func
from modules_core.manager_voiceid_v3 import ManagerVoiceIdV3

# pytorch model process (shared accross sessions because of limited GPU resources)

FREE = 0
NOT_PROCESSED = 1
DONE = 2

class WorkerEmbeddingsExtractorV3(Process):

    def __init__(self, args, cpu_affinity=None, cuda_device=0):
        self.cpu_affinity = cpu_affinity
        self.cuda_device = cuda_device
        if sys.platform != 'darwin':
            if multiprocessing.get_start_method() != 'spawn':
                LoggingUtils.info('setting set_start_method spawn')
                multiprocessing.set_start_method('spawn', force=True)
        super().__init__(daemon=False)

        # coppied params
        self.args = args

        self.daemon = False # Not allowed to use with join together (if not listed it will inherit this from parent)

        self.is_running = RawValue(ctypes.c_bool, True)

        self.max_data_input_frames = 64
        self.len_data_input_frame = self.args.input_timesteps_max * self.args.nb_sfft_bins
        self.max_data_input = self.max_data_input_frames * self.len_data_input_frame
        self.max_data_output = self.max_data_input_frames * self.args.embedding_size
        self.max_data_output_emo = self.max_data_input_frames * len(self.args.emotions)

        self.gender_model_percent = RawValue(ctypes.c_float, 0.5)

        self.array_data_input_spec = RawArray(ctypes.c_float, self.max_data_input)
        self.array_data_input_model_iteration_id = RawArray(ctypes.c_uint32, self.max_data_input_frames)
        self.array_data_input_is_male = RawArray(ctypes.c_bool, self.max_data_input_frames)

        # data output - id vectors (flattened 1d)

        self.array_data_slot_state = RawArray(ctypes.c_byte, self.max_data_input_frames)
        self.lock_data_slot_state = Lock()

        self.array_data_output_vec = RawArray(ctypes.c_float, self.max_data_output)
        self.array_data_output_emo = RawArray(ctypes.c_float, self.max_data_output_emo)

        self.array_data_output_male_percent = RawArray(ctypes.c_float, self.max_data_input_frames)
        self.array_data_output_laugh_percent = RawArray(ctypes.c_float, self.max_data_input_frames)

        self.is_started = RawValue(ctypes.c_bool, False)
        self.is_signal_to_log_status = RawValue(ctypes.c_bool, False)

        self.voiceid_male_model_iteration_id = 0
        self.voiceid_female_model_iteration_id = 0

        self.model_female: torch.nn.Module  = None
        self.model_male: torch.nn.Module = None

        self.last_voiceid_male_model_iteration_id = 0
        self.last_voiceid_female_model_iteration_id = 0

        self.state_dict_female = None
        self.state_dict_male = None

    def start(self):
        super().start()
        while not self.is_started.value:
            time.sleep(0.01)

    @staticmethod
    def send_sync(
            workers_diarization_extractors,
            frame_spec,
            model_iteration_id,
            is_male
    ):
        slot_id = None
        is_running = True
        worker = None
        while is_running:
            for worker in workers_diarization_extractors:
                is_running = worker.is_running.value
                slot_id = worker.set_input_data(
                    frame_spec,
                    model_iteration_id,
                    is_male
                )
                if slot_id is not None:
                    break
            if slot_id is not None:
                break
            else:
                time.sleep(1e-3)

        output = None
        while is_running:
            is_running = worker.is_running.value
            tmp = worker.get_data_output([slot_id])
            if len(tmp) and len(tmp[0]):
                frames_vec, frames_emo, male_percents, laugh_percents = tmp
                output = frames_vec[0], frames_emo[0], male_percents[0], laugh_percents[0]
                break
            else:
                time.sleep(1e-3)
        return output

    # shared function
    # if no slots available then return None
    def set_input_data(
            self,
            frame_spec,
            model_iteration_id,
            is_male
    ):
        slot_id = None

        try:
            for idx_slot in range(self.max_data_input_frames):
                self.lock_data_slot_state.acquire()
                if self.array_data_slot_state[idx_slot] == FREE:
                    slot_id = idx_slot

                    idx_input_start = idx_slot * self.len_data_input_frame
                    idx_input_end = (idx_slot + 1) * self.len_data_input_frame

                    self.array_data_input_spec[idx_input_start:idx_input_end] = frame_spec.flatten()[:]
                    self.array_data_input_model_iteration_id[idx_slot] = model_iteration_id
                    self.array_data_input_is_male[idx_slot] = is_male

                    self.array_data_slot_state[idx_slot] = NOT_PROCESSED
                try:
                    self.lock_data_slot_state.release()
                except:
                    pass
                if slot_id is not None:
                    break
        except Exception as e:
            LoggingUtils.exception(e)

        return slot_id

    # shared function
    # return embedding vectors // if not ready return empty vectors
    def get_data_output(self, slot_ids):
        frames_vec = []
        frames_emo = []
        male_percents = []
        laugh_percents = []

        try:
            is_all_slots_completed = True
            for idx_slot in slot_ids:
                if self.array_data_slot_state[idx_slot] != DONE:
                    is_all_slots_completed = False
                    break

            if is_all_slots_completed:
                for idx_slot in slot_ids:
                    if self.array_data_slot_state[idx_slot] == DONE:

                        vec = self.array_data_output_vec[idx_slot*self.args.embedding_size:(idx_slot+1)*self.args.embedding_size]
                        vec = np.array(vec)

                        emo = self.array_data_output_emo[idx_slot*len(self.args.emotions):(idx_slot+1)*len(self.args.emotions)]
                        emo = np.array(emo)

                        male_percent = self.array_data_output_male_percent[idx_slot]
                        laugh_percent = self.array_data_output_laugh_percent[idx_slot]

                        frames_emo.append(emo)
                        frames_vec.append(vec)
                        male_percents.append(male_percent)
                        laugh_percents.append(laugh_percent)

                        # free slot
                        self.array_data_slot_state[idx_slot] = FREE

        except Exception as e:
            LoggingUtils.error(str(e))
            exc_type, exc_value, exc_tb = sys.exc_info()
            LoggingUtils.error(traceback.format_exception(exc_type, exc_value, exc_tb))

        return frames_vec, frames_emo, male_percents, laugh_percents

    # shared function
    def stop(self):
        self.is_running.value = False

    def load_voiceid_model(self, voiceid_male_model_iteration_id, voiceid_female_model_iteration_id):
        try:

            if voiceid_male_model_iteration_id != self.voiceid_male_model_iteration_id:
                LoggingUtils.info(f'updating male model iteration: {self.voiceid_male_model_iteration_id} to {voiceid_male_model_iteration_id}')
                self.model_male, _ = self.load_model('Model', ManagerVoiceIdV3.get_model_iteration_path(voiceid_male_model_iteration_id))
                self.state_dict_male = self.model_male.state_dict()

            # based not on iterations, but on same model core
            if self.args.voiceid_male_model_id == self.args.voiceid_female_model_id:
                self.model_female = self.model_male
                self.state_dict_female = self.state_dict_male
            else:
                if voiceid_female_model_iteration_id != self.voiceid_female_model_iteration_id:
                    LoggingUtils.info(f'updating female model iteration: {self.voiceid_female_model_iteration_id} to {voiceid_female_model_iteration_id}')
                    self.model_female, _ = self.load_model('Model', ManagerVoiceIdV3.get_model_iteration_path(voiceid_female_model_iteration_id))
                    self.state_dict_female = self.model_female.state_dict()

            self.voiceid_male_model_iteration_id = voiceid_male_model_iteration_id
            self.voiceid_female_model_iteration_id = voiceid_female_model_iteration_id
        except Exception as e:
            LoggingUtils.exception(e)

    def initialization(self):
        try:
            self.lock_data_slot_state.acquire()
            try:
                for idx_slot in range(self.max_data_input_frames):
                    self.array_data_slot_state[idx_slot] = FREE
                    self.array_data_output_male_percent[idx_slot] = 0.0
                    self.array_data_output_laugh_percent[idx_slot] = 0.0
            except Exception as e:
                LoggingUtils.exception(e)
            self.lock_data_slot_state.release()

            self.load_voiceid_model(
                 ManagerVoiceIdV3.get_latest_model_iteration_id(self.args.voiceid_male_model_id, user_id=None),
                 ManagerVoiceIdV3.get_latest_model_iteration_id(self.args.voiceid_female_model_id, user_id=None)
            )

            self.model_gender, self.model_gender_data_type = self.load_model('ModelBinary', self.args.path_model_gender_args)

            self.model_laugh = None
            self.model_laugh_data_type = ''
            if os.path.exists(self.args.path_model_laugh_args):
                self.model_laugh, self.model_laugh_data_type = self.load_model('ModelLaugh', self.args.path_model_laugh_args)

            self.model_emo_male, self.model_emo_male_data_type = self.load_model('ModelEmo', self.args.path_model_emo_male_args)
            if self.args.path_model_emo_male_args != self.args.path_model_emo_female_args:
                self.model_emo_female, self.model_emo_female_data_type = self.load_model('ModelEmo', self.args.path_model_emo_female_args)

        except Exception as e:
            LoggingUtils.exception(e)

    def load_model(self, model_name: str, path_model_args: str) -> Tuple[torch.nn.Module, str]:
        LoggingUtils.info(f'loading: {model_name} {path_model_args}')

        path_model = path_model_args.replace('.json', '.pt')

        with open(path_model_args, 'r') as fp:
            args = json.load(fp)
            args = DictToObj(**args)
        args.tf_path_data_args_json = self.args.path_data_args
        args.device = 'cuda'
        args.is_cuda = True
        data_type = args.tf_data_type

        if hasattr(args, 'model'):
            model_name = args.model
        else:
            model_name = args.tf_model

        Model = getattr(__import__('modules_ext.' + model_name, fromlist=[model_name]), 'Model')
        model = Model(args)

        if not torch.cuda.is_available():
            self.args.device = 'cpu'
            LoggingUtils.info(f'{model_name} cuda not loaded')

        if self.args.device == 'cuda':
            model = model.cuda(self.cuda_device)
            LoggingUtils.info(f'{model_name} cuda loaded')

        LoggingUtils.info(f'{model_name} params from file load started')

        model = model.eval()

        if self.args.device == 'cuda':
            state_dict = torch.load(path_model)
        else:
            state_dict = torch.load(path_model, map_location='cpu')

        state_dict_fix = OrderedDict({})
        for key, value in state_dict.items():
            state_dict_fix[key.replace('module.', '')] = value
        model.load_state_dict(state_dict_fix)

        LoggingUtils.info(f'{model_name} loaded')

        #if self.args.is_data_parallel:
        #    model_speech = model_speech.module

        model = model.eval()  # very important call to avoid dropouts

        return model, data_type

    def convert_output(self, t_output):
        if self.args.device == 'cuda':
            t_output = t_output.cpu()
        np_output = t_output.data.numpy()
        return np_output

    def scale_sub_frames(self, sub_frames_input):
        sub_frames = sub_frames_input[:]
        data_min = np.min(sub_frames)
        data_max = np.max(sub_frames)
        if data_max != data_min:
            sub_frames -= data_min
            sub_frames /= (data_max - data_min)
        return sub_frames

    def pre_process_data(self, sub_frames, args):
        if len(sub_frames) > 0:
            if args.is_data_scaled_by_frame:
                if hasattr(args, 'data_pre_spec_y_min'):
                    sub_frames[:, args.data_pre_spec_y_min] = 0.0
                    sub_frames[:, min(sub_frames.shape[1], args.data_pre_spec_y_max):] = 0.0
                if hasattr(args, 'data_pre_process_power'):
                    sub_frames = sub_frames ** args.data_pre_process_power
                data_min = np.min(sub_frames)
                data_max = np.max(sub_frames)
                if data_min != 0 and data_max != 1:
                    if data_max != data_min:
                        sub_frames -= data_min
                        sub_frames /= (data_max - data_min)

            if hasattr(args, 'data_pre_threshold'):
                if args.data_pre_threshold > 0:
                    sub_frames[sub_frames < args.data_pre_threshold] = 0.0
        return sub_frames

    def convert_frame_to_tensor(self, np_frame, model):
        t_frame_spec = torch.FloatTensor(
            self.pre_process_data(np_frame.astype(np.float), model.args)
        )
        if self.args.device == 'cuda':
            t_frame_spec = t_frame_spec.cuda(self.cuda_device)
        t_frame_spec = torch.unsqueeze(t_frame_spec, dim=0)
        return t_frame_spec


    def set_signal_to_log(self):
        self.is_signal_to_log_status.value = True

    def run(self):
        profile_func(self.run_inner, self.args, profile_name='diarization_extractor')

    def run_inner(self):
        LoggingUtils.remove_handlers()
        LoggingUtils.create()

        p = psutil.Process()
        all_cpus = np.arange(psutil.cpu_count(), dtype=np.int).tolist()
        p.cpu_affinity(all_cpus)
        torch.set_num_threads(psutil.cpu_count())
        LoggingUtils.warn(f'Diarization: {p.pid} cpu_affinity: {p.cpu_affinity()}')

        self.voiceid_male_model_iteration_id = 0
        self.voiceid_female_model_iteration_id = 0

        self.initialization()

        original_sigint_handler = signal.getsignal(signal.SIGINT)
        original_sigterm_handler = signal.getsignal(signal.SIGTERM)

        def func_terminate(signum, frame):
            signal.signal(signal.SIGINT, original_sigint_handler)
            signal.signal(signal.SIGTERM, original_sigterm_handler)
            LoggingUtils.info('terminate signal diarization_extractor')
            self.is_running.value = False
            try:
                self.lock_data_slot_state.release()
            except:
                pass

        signal.signal(signal.SIGINT, func_terminate)
        signal.signal(signal.SIGTERM, func_terminate)

        try:

            self.is_started.value = True
            time_last_activity = time.time()

            with torch.no_grad():
                while self.is_running.value:

                    time.sleep(1e-3) # very important because of GIL
                    if self.is_signal_to_log_status:
                        self.is_signal_to_log_status = False
                        log_process_id('diarization_extractor', 'signal', self.args)

                    try:
                        for idx_slot in range(self.max_data_input_frames):
                            if self.array_data_slot_state[idx_slot] == NOT_PROCESSED:

                                np_frame_spec = np.array(self.array_data_input_spec[idx_slot*self.len_data_input_frame: (idx_slot+1)*self.len_data_input_frame]).astype(np.float)
                                np_frame_spec = np.reshape(np_frame_spec, (self.args.input_timesteps_max, self.args.nb_sfft_bins))

                                # only available for conversation, not voiceID
                                target_model_iteration_id = self.array_data_input_model_iteration_id[idx_slot]
                                target_is_male = self.array_data_input_is_male[idx_slot]

                                np_gender_output = np.zeros((2,))
                                np_laugh_output = np.zeros((2,))
                                np_emo_output = np.zeros((len(self.args.emotions), ))
                                np_speech_output = np.zeros((self.model_male.args.embedding_size, ))


                                try:
                                    #automate process

                                    #TODO paralelize models
                                    t_frame_spec = self.convert_frame_to_tensor(np_frame_spec, self.model_gender)
                                    t_gender_output = self.model_gender.forward_runtime(
                                        t_frame_spec,
                                        torch.zeros_like(t_frame_spec))
                                    np_gender_output = self.convert_output(t_gender_output[0])

                                    t_frame_spec = self.convert_frame_to_tensor(np_frame_spec, self.model_laugh)
                                    t_laugh_output = self.model_laugh.forward_runtime(
                                        t_frame_spec,
                                        torch.zeros_like(t_frame_spec))
                                    np_laugh_output = self.convert_output(t_laugh_output[0])

                                    is_male = True
                                    model_id = self.args.voiceid_male_model_id
                                    state_dict_global = self.state_dict_male
                                    model_speech = self.model_male
                                    model_iteration_id_global = self.voiceid_male_model_iteration_id
                                    model_emo = self.model_emo_male

                                    if np_gender_output[0] < self.gender_model_percent.value:
                                        is_male = False
                                        state_dict_global = self.state_dict_female
                                        model_id = self.args.voiceid_female_model_id
                                        model_speech = self.model_female
                                        model_iteration_id_global = self.voiceid_female_model_iteration_id
                                        model_emo = self.model_emo_female

                                    # opposite gender should use generic base model
                                    if target_is_male != is_male or target_model_iteration_id == 0:
                                        target_model_iteration_id = model_iteration_id_global

                                    is_to_load = False
                                    if target_is_male == is_male:
                                        if is_male:
                                            if self.last_voiceid_male_model_iteration_id != target_model_iteration_id:
                                                if target_model_iteration_id != 0 or self.last_voiceid_male_model_iteration_id != self.voiceid_male_model_iteration_id:
                                                    is_to_load = True
                                        else:
                                            if self.last_voiceid_female_model_iteration_id != target_model_iteration_id:
                                                if target_model_iteration_id != 0 or self.last_voiceid_male_model_iteration_id != self.voiceid_female_model_iteration_id:
                                                    is_to_load = True

                                    if is_to_load:
                                        try:
                                            is_loaded = False
                                            if target_model_iteration_id != 0:
                                                state_dict_custom = ManagerVoiceIdV3.get_model_iteration_state_dict(
                                                    target_model_iteration_id
                                                )
                                                if state_dict_custom is not None:
                                                    LoggingUtils.debug(f'state_dict_custom last_model_iteration_id: {target_model_iteration_id}')
                                                    model_speech.load_state_dict(pickle.loads(state_dict_custom))
                                                    LoggingUtils.debug('state_dict_custom loaded')
                                                    is_loaded = True
                                                else:
                                                    LoggingUtils.debug(f'state_dict_custom missing last_model_iteration_id: {target_model_iteration_id}')

                                            if not is_loaded:
                                                LoggingUtils.debug('state_dict_custom state_dict_global')
                                                target_model_iteration_id = model_iteration_id_global
                                                model_speech.load_state_dict(state_dict_global)
                                        except Exception as exc:
                                            LoggingUtils.exception(exc)

                                        if is_male:
                                            self.last_voiceid_male_model_iteration_id = target_model_iteration_id
                                        else:
                                            self.last_voiceid_female_model_iteration_id = target_model_iteration_id

                                    #TODO paralelize models (loading voiceId as well)
                                    t_frame_spec = self.convert_frame_to_tensor(np_frame_spec, model_speech)
                                    t_speech_output, _ = model_speech.forward(t_frame_spec)
                                    np_speech_output = self.convert_output(t_speech_output[0])

                                    t_frame_spec = self.convert_frame_to_tensor(np_frame_spec, model_emo)
                                    t_emo_output = model_emo.forward_runtime(t_frame_spec)
                                    np_emo_output = self.convert_output(t_emo_output[0])

                                except Exception as e:
                                    LoggingUtils.exception(e)

                                self.array_data_output_male_percent[idx_slot] = np_gender_output[0]
                                self.array_data_output_laugh_percent[idx_slot] = np_laugh_output[0]
                                self.array_data_output_emo[idx_slot*len(self.args.emotions):(idx_slot+1)*len(self.args.emotions)] = np_emo_output.flatten().tolist()
                                self.array_data_output_vec[idx_slot*self.args.speaker_embedding_size:(idx_slot+1)*self.args.speaker_embedding_size] = np_speech_output.flatten().tolist()
                                self.array_data_slot_state[idx_slot] = DONE

                    except Exception as e:
                        LoggingUtils.exception(e)

        except Exception as e:
            LoggingUtils.exception(e)
        log_process_id('diarization_extractor', 'signal', self.args)
        LoggingUtils.warn('exit diarization_extractor.run')